/**
 * DESCRIPTION OF THE FILE
 *
 * @author Michal Kravčenko
 * @date 30.7.18 -
 */

#ifndef INC_4NEURO_GRADIENTDESCENT_H
#define INC_4NEURO_GRADIENTDESCENT_H

#include "../settings.h"
#include "../constants.h"
#include "ILearningMethods.h"
#include "../ErrorFunction/ErrorFunctions.h"

namespace lib4neuro {
    /**
     *
     */
    class GradientDescent : public ILearningMethods {

    private:

        /**
         * Threshold for the successful ending of the optimization - deviation from minima
         */
        double tolerance;

        /**
         *
         */
        double max_error;

        /**
         * Number of iterations to reset step size to tolerance/10.0
         */
        size_t restart_frequency;

        /**
         *
         */
		size_t batch;

		/**
		 *
		 */
		size_t iter_max;

        /**
         * Maximal number of iterations - optimization will stop after that, even if not converged
         */
        long long int maximum_niters;

        /**
         * Vector of minima coordinates
         */
        std::vector<double> *optimal_parameters;

        /**
         * Adaptive calculation of the step-size based on several historical characteristics.
         * ----------------------------------------------------------------------------------
         * If the current error @fi is larger than the error in the previous step @fim, the rate of step-size change decreases (the algorithm is going in the direction too quickly)
         * Otherwise the rate of step-size change increases (the algorithm is on the right path, we can attempts to push through more rapidly)
         * ----------------------------------------------------------------------------------
         * The step size is then calculated via: @c^(1-2@beta) * (@grad_norm_prev/@grad_norm)^(1/@c)
         * If the previous gradient norm is lower then the current gradient norm, then the step-size decreases (as we probably took a too large of a step)
         * Otherwise it increases (as we are likely on the right track, we can try to speed-up the convergence)
         *
         * @param gamma[in, out] a step size used in the last iteration
         * @param beta[in] a number in the interval [0, 1]. it represents a measure of direction change between two last steps, 0: no change, 1:opposite directions
         * @param c[in, out] greater than zero. it is a measure of a non-linear step-size change. the higher @c is, the more rapidly the step-size increases/decreases
         * @param grad_norm_prev[in] gradient norm of the error in the previous iteration
         * @param grad_norm[in] gradient norm of the error in the current iteration
         * @param fi[in] value of the error
         * @param fim[in] value of the error in the previous iteration
         */
        virtual void
        eval_step_size_mk(double &gamma,
                          double beta,
                          double &c,
                          double grad_norm_prev,
                          double grad_norm,
                          double fi,
                          double fim);

    public:

        /**
         * Creates an instance of Gradient Descent Optimizer (i.e. back-propagation)
         * @param epsilon Threshold for the successful ending of the optimization - deviation from minima
         * @param n_to_restart Number of iterations to reset step size to tolerance/10.0
         * @param max_iters Maximal number of iterations - optimization will stop after that, even if not converged
         */
        LIB4NEURO_API GradientDescent(double epsilon = 1e-3, size_t n_to_restart = 100, int max_iters = 1000, size_t batch = 0);

        /**
         * Deallocates the instance
         */
        LIB4NEURO_API ~GradientDescent();

        /**
         *
         * @param ef
         */
        LIB4NEURO_API virtual void optimize(lib4neuro::ErrorFunction &ef);

        /**
         *
         * @return
         */
        LIB4NEURO_API virtual std::vector<double> *get_parameters();
    };

}

#endif //INC_4NEURO_GRADIENTDESCENT_H
